/*
 * gps-manager
 *
 * Copyright (c) 2011-2013 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: Youngae Kang <youngae.kang@samsung.com>, Minjune Kim <sena06.kim@samsung.com>
 *          Genie Kim <daejins.kim@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <config.h>

#include <string.h>
#include <time.h>
#include <stdlib.h>

#include <glib.h>
#include <glib-object.h>

#include <lbs_dbus_server.h>

#include "gps_manager_data_types.h"
#include "gps_manager.h"

#include "server.h"
#include "last_position.h"
#include "debug_util.h"

#define GPSMANAGER_DBUS_SERVICE	"org.tizen.lbs.Providers.GpsManager"
#define GPSMANAGER_DBUS_PATH	"/org/tizen/lbs/Providers/GpsManager"

typedef struct {
	GMainLoop *loop;
	GMutex *mutex;

	LbsStatus status;
	pos_data_t position;
	sv_data_t satellite;
	nmea_data_t nmea;

	gboolean is_running;
	gint client_count;
	lbs_server_dbus_h lbs_server;
} GpsManager;

static void start_tracking(GpsManager * gpsmanager)
{
	LOG_GPS(DBG_LOW, "start_tracking");
	gpsmanager->status = LBS_STATUS_ACQUIRING;
	g_mutex_lock(gpsmanager->mutex);
	if (request_start_session() == TRUE) {
		gpsmanager->is_running = TRUE;
	} else {
		LOG_GPS(DBG_ERR, "Fail to request_start_session");
	}
	g_mutex_unlock(gpsmanager->mutex);
}

static void stop_tracking(GpsManager * gpsmanager)
{
	LOG_GPS(DBG_LOW, "stop_tracking");
	gps_manager_set_last_position(&gpsmanager->position);

	g_mutex_lock(gpsmanager->mutex);
	if (request_stop_session() == TRUE) {
		gpsmanager->is_running = FALSE;
	}
	g_mutex_unlock(gpsmanager->mutex);
	gpsmanager->status = LBS_STATUS_UNAVAILABLE;

	lbs_server_emit_status_changed(gpsmanager->lbs_server, LBS_STATUS_UNAVAILABLE);
}

static void set_options(GVariant *options, gpointer userdata)
{
	LOG_GPS(DBG_LOW, "set_options callback");

	GpsManager *gpsmanager = (GpsManager *)userdata;

	GVariantIter iter;
	gchar *key = NULL;
	GVariant *value = NULL;
	gsize length = 0;

	g_variant_iter_init(&iter, options);
	g_variant_iter_next(&iter, "{&sv}", &key, &value);
	LOG_GPS(DBG_LOW, "set_options key[%s]", key);
	if (!g_strcmp0(key, "CMD")) {
		if (!g_strcmp0(g_variant_get_string(value, &length), "START")) {
			g_mutex_lock(gpsmanager->mutex);
			gpsmanager->client_count++;
			g_mutex_unlock(gpsmanager->mutex);

			if (gpsmanager->is_running == TRUE) {
				LOG_GPS(DBG_LOW, "gps-manager is already running");
				return;
			}
			start_tracking(gpsmanager);
		} else if (!g_strcmp0(g_variant_get_string(value, &length), "STOP")) {
			g_mutex_lock(gpsmanager->mutex);
			gpsmanager->client_count--;
			g_mutex_unlock(gpsmanager->mutex);

			if (gpsmanager->is_running == FALSE) {
				return;
			}
			if (gpsmanager->client_count <= 0 ) {
				stop_tracking(gpsmanager);
			}
		}
	}
}

static gboolean remove_all_clients(GpsManager *gpsmanager)
{
	LOG_GPS(DBG_LOW, "remove_all_clients");
	if (gpsmanager->client_count <= 0) {
		return FALSE;
	}

	gpsmanager->client_count = 0;
	stop_tracking(gpsmanager);

	return TRUE;
}

static void shutdown(gpointer userdata)
{
	LOG_GPS(DBG_LOW, "shutdown callback");
	GpsManager *gpsmanager = (GpsManager *)userdata;

	if (gpsmanager->is_running) {
		if (remove_all_clients(gpsmanager)) {
			LOG_GPS(DBG_ERR, "<<<< Abnormal shutdown >>>>");
		}
	}
}

static void update_position_cb(pos_data_t *pos, gps_error_t error, void *user_data)
{
	LOG_GPS(DBG_LOW, "update_position_cb");

	GVariant *accuracy = NULL;
	LbsPositionExtFields fields;

	GpsManager *gpsmanager = (GpsManager *)(user_data);

	memcpy(&gpsmanager->position, pos, sizeof(pos_data_t));

	if (gpsmanager->status != LBS_STATUS_AVAILABLE) {
		lbs_server_emit_status_changed(gpsmanager->lbs_server, LBS_STATUS_AVAILABLE);
		gpsmanager->status = LBS_STATUS_AVAILABLE;
	}

	fields = (LBS_POSITION_EXT_FIELDS_LATITUDE | LBS_POSITION_EXT_FIELDS_LONGITUDE | LBS_POSITION_EXT_FIELDS_ALTITUDE |
			LBS_POSITION_EXT_FIELDS_SPEED | LBS_POSITION_EXT_FIELDS_DIRECTION);

	accuracy = g_variant_new_parsed("(%i, %d, %d)", LBS_ACCURACY_LEVEL_DETAILED, pos->hor_accuracy, pos->ver_accuracy);
	lbs_server_emit_position_changed(gpsmanager->lbs_server,
					fields, pos->timestamp, pos->latitude, pos->longitude, pos->altitude,
						pos->speed, pos->bearing, 0.0, accuracy);
}

static void update_satellite_cb(sv_data_t * sv, void *user_data)
{
	LOG_GPS(DBG_LOW, "update_satellite_cb");
	int index;
	int timestamp = 0;
	int satellite_used = 0;
	GVariant *used_prn = NULL;
	GVariantBuilder *used_prn_builder = NULL;
	GVariant *satellite_info = NULL;
	GVariantBuilder *satellite_info_builder = NULL;

	GpsManager *gpsmanager = (GpsManager *)(user_data);

	memcpy(&gpsmanager->satellite, sv, sizeof(sv_data_t));
	timestamp = sv->timestamp;

	used_prn_builder = g_variant_builder_new(G_VARIANT_TYPE ("ai"));
	for (index = 0; index < sv->num_of_sat; ++index) {
		if (sv->sat[index].used) {
			g_variant_builder_add(used_prn_builder, "i", sv->sat[index].prn);
			++satellite_used;
		}
	}
	used_prn = g_variant_builder_end(used_prn_builder);

	satellite_info_builder = g_variant_builder_new(G_VARIANT_TYPE ("a(iiii)"));
	for (index = 0; index < sv->num_of_sat; ++index) {
		g_variant_builder_add(satellite_info_builder, "(iiii)",  sv->sat[index].prn,
				 sv->sat[index].elevation, sv->sat[index].azimuth, sv->sat[index].snr);
	}
	satellite_info = g_variant_builder_end(satellite_info_builder);

	lbs_server_emit_satellite_changed(gpsmanager->lbs_server, timestamp, satellite_used, sv->num_of_sat,
							used_prn, satellite_info);
}

static void update_nmea_cb(nmea_data_t * nmea, void *user_data)
{
	int timestamp = 0;

	GpsManager *gpsmanager = (GpsManager *)(user_data);

	if (gpsmanager->nmea.data) {
		g_free(gpsmanager->nmea.data);
		gpsmanager->nmea.len = 0;
		gpsmanager->nmea.data = NULL;
	}
	gpsmanager->nmea.data = g_malloc(nmea->len + 1);
	g_memmove(gpsmanager->nmea.data, nmea->data, nmea->len);
	gpsmanager->nmea.data[nmea->len] = '\0';
	gpsmanager->nmea.len = nmea->len;

	timestamp = gpsmanager->position.timestamp;

	lbs_server_emit_nmea_changed(gpsmanager->lbs_server, timestamp, gpsmanager->nmea.data);
}

static void gpsmanager_init(GpsManager * gpsmanager)
{
	LOG_GPS(DBG_LOW, "gpsmanager_init");

	gpsmanager->status = LBS_STATUS_UNAVAILABLE;
	gpsmanager->mutex = g_mutex_new();

	memset(&gpsmanager->position, 0x00, sizeof(pos_data_t));
	memset(&gpsmanager->satellite, 0x00, sizeof(sv_data_t));
	memset(&gpsmanager->nmea, 0x00, sizeof(nmea_data_t));

	gpsmanager->is_running = FALSE;
	gpsmanager->client_count = 0;
}

int main(int argc, char **argv)
{
	GpsManager *gpsmanager = NULL;
	struct gps_callbacks cb;
	cb.pos_cb = update_position_cb;
	cb.sv_cb = update_satellite_cb;
	cb.nmea_cb = update_nmea_cb;

#if !GLIB_CHECK_VERSION (2, 31, 0)
	if (!g_thread_supported()) {
		g_thread_init(NULL);
	}
#endif

	g_type_init();

	initialize_server(argc, argv);

	gpsmanager = g_new0(GpsManager, 1);
	if (!gpsmanager) {
		LOG_GPS(DBG_ERR, "GpsManager create fail");
		return 1;
	}
	gpsmanager_init(gpsmanager);

	register_update_callbacks(&cb, gpsmanager);

	/* create lbs-dbus server */
	lbs_server_create(GPSMANAGER_DBUS_SERVICE, GPSMANAGER_DBUS_PATH, "gps-manager", "GPS Manager",
		&(gpsmanager->lbs_server), set_options, shutdown, (void *)gpsmanager);
	LOG_GPS(DBG_LOW, "lbs_server_create called");

	gpsmanager->loop = g_main_loop_new(NULL, TRUE);
	g_main_loop_run(gpsmanager->loop);

	LOG_GPS(DBG_LOW, "GPS_manager deamon Stop....");
	/* destroy lbs-dbus server */
	lbs_server_destroy(gpsmanager->lbs_server);
	LOG_GPS(DBG_LOW, "lbs_server_destroy called");

	g_main_loop_unref(gpsmanager->loop);

	g_free(gpsmanager);

	deinitialize_server();

	return 0;
}
