/*
 * gps-manager
 *
 * Copyright (c) 2011-2013 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: Youngae Kang <youngae.kang@samsung.com>, Minjune Kim <sena06.kim@samsung.com>
 *          Genie Kim <daejins.kim@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <glib.h>

#include <vconf.h>
#include <vconf-internal-location-keys.h>
#include <location-module.h>

#include <dlfcn.h>
#include <lbs_dbus_client.h>


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "log.h"

typedef struct {
	char devname[256];
	lbs_client_dbus_h lbs_client;
	LocModStatusCB status_cb;
	LocModPositionExtCB pos_cb;
	LocModSatelliteCB sat_cb;
	gpointer userdata;
	gboolean is_started;
} GpsManagerData;

#define GPS_MANAGER_SERVICE_NAME	"org.tizen.lbs.Providers.GpsManager"
#define GPS_MANAGER_SERVICE_PATH	"/org/tizen/lbs/Providers/GpsManager"

#define MPS_TO_KMPH 	3.6 	// 1 m/s = 3.6 km/h

static void status_callback(GVariant *param, void *user_data)
{
	GpsManagerData *gps_manager = (GpsManagerData *) user_data;
	g_return_if_fail(param);
	g_return_if_fail(gps_manager);
	g_return_if_fail(gps_manager->status_cb);

	int status = 0;

	g_variant_get (param, "(i)", &status);

	if (status == 3) {	//TODO:  LBS_STATUS_AVAILABLE ?
		MOD_LOGD("LBS_STATUS_AVAILABLE");
		gps_manager->status_cb(TRUE, LOCATION_STATUS_3D_FIX, gps_manager->userdata);
	}
	else {
		MOD_LOGD("LBS_STATUS_ACQUIRING/ERROR/UNAVAILABLE. Status[%d]", status);
		gps_manager->status_cb(FALSE, LOCATION_STATUS_NO_FIX, gps_manager->userdata);
	}
}

static void satellite_callback(GVariant *param, void *user_data)
{
	GpsManagerData *gps_manager = (GpsManagerData *)user_data;
	g_return_if_fail(gps_manager);
	g_return_if_fail(gps_manager->sat_cb);

	guint idx;
	guint used_idx;
	guint *used_prn_array = NULL;
	gboolean ret = FALSE;
	int timestamp = 0, satellite_used = 0, satellite_visible = 0;

	LocationSatellite *sat = NULL;
	GVariant *used_prn = NULL;
	GVariantIter *used_prn_iter = NULL;
	GVariant *sat_info = NULL;
	GVariantIter *sat_iter = NULL;
	int prn = 0, elev = 0, azim = 0, snr = 0;

	g_variant_get(param, "(iii@ai@a(iiii))", &timestamp, &satellite_used, &satellite_visible, &used_prn, &sat_info);
	g_variant_get(used_prn, "ai", &used_prn_iter);
	g_variant_get(sat_info, "a(iiii)", &sat_iter);
	MOD_LOGD("timestamp [%d], satellite_used [%d], satellite_visible[%d]", timestamp, satellite_used, satellite_visible);
	int tmp_prn = 0;
	int num_of_used_prn = g_variant_iter_n_children (used_prn_iter);
	if (num_of_used_prn > 0) {
		used_prn_array = (guint *)g_new0(guint, num_of_used_prn);
		for (idx = 0; idx < num_of_used_prn; idx++) {
			ret = g_variant_iter_next(used_prn_iter, "i", &tmp_prn);
			if (ret == FALSE)
				break;
			used_prn_array[idx] = tmp_prn;
		}
	}

	sat = location_satellite_new(satellite_visible);

	sat->timestamp = timestamp;
	sat->num_of_sat_inview = satellite_visible;
	sat->num_of_sat_used = satellite_used;

	GVariant *tmp_var = NULL;
	for (idx = 0; idx < satellite_visible; idx++) {
		gboolean used = FALSE;
		tmp_var = g_variant_iter_next_value (sat_iter);
		g_variant_get (tmp_var, "(iiii)",  &prn, &elev, &azim, &snr);
		if (used_prn_array != NULL) {
			for (used_idx = 0; used_idx < satellite_used; used_idx++) {
				if (prn == used_prn_array[used_idx]) {
					used = TRUE;
					break;
				}
			}
		}
		MOD_LOGD("prn[%d] : used %d elev %d azim %d snr %d", prn, used, elev, azim, snr);
		location_satellite_set_satellite_details(sat, idx, prn, used, elev, azim, snr);
	}

	gps_manager->sat_cb(TRUE, sat, gps_manager->userdata);
	location_satellite_free(sat);

	if (used_prn_array) {
		g_free(used_prn_array);
		used_prn_array = NULL;
	}
}

#if 0
static void nmea_callback(GeoclueNmea * nmea, int timestamp, char *data, gpointer userdata)
{
}
#endif

static void position_callback(GVariant *param, void *user_data)
{
	MOD_LOGD("position_callback");
	GpsManagerData *gps_manager = (GpsManagerData *)user_data;
	g_return_if_fail(gps_manager);
	g_return_if_fail(gps_manager->pos_cb);

	int fields = 0 ,timestamp = 0 , level = 0;
	double latitude = 0.0, longitude = 0.0, altitude = 0.0, speed = 0.0, direction = 0.0, climb = 0.0, horizontal = 0.0, vertical = 0.0;
	GVariant *accuracy = NULL;

	g_variant_get(param, "(iidddddd@(idd))", &fields, &timestamp, &latitude, &longitude, &altitude, &speed, &direction, &climb, &accuracy);
	g_variant_get(accuracy, "(idd)", &level, &horizontal, &vertical);

	LocationPosition *pos = NULL;
	LocationVelocity *vel = NULL;
	LocationAccuracy *acc = NULL;

	if (altitude) {
		pos = location_position_new(timestamp, latitude, longitude, altitude, LOCATION_STATUS_3D_FIX);
	} else {
		pos = location_position_new(timestamp, latitude, longitude, 0.0, LOCATION_STATUS_2D_FIX);
	}

	vel = location_velocity_new(timestamp, speed, direction, climb);

	acc = location_accuracy_new(LOCATION_ACCURACY_LEVEL_DETAILED, horizontal, vertical);

	MOD_LOGD
	    ("time(%d) lat(%f) long(%f) alt(%f) status(%d) acc_level(%d) hoz_acc(%f) vert_acc(%f)",
	     pos->timestamp, pos->latitude, pos->longitude, pos->altitude,
	     pos->status, acc->level, acc->horizontal_accuracy, acc->vertical_accuracy);

	gps_manager->pos_cb(TRUE, pos, vel, acc, gps_manager->userdata);

	location_position_free(pos);
	location_velocity_free(vel);
	location_accuracy_free(acc);
}

static void on_signal_callback(GDBusConnection *conn, const gchar *name, const gchar *path, const gchar *interface, const gchar *sig, GVariant *param, gpointer user_data)
{
	MOD_LOGD("name : %s, path : %s, interface : %s, sig : %s\n", name, path, interface, sig);

	if (!g_strcmp0(sig, "SatelliteChanged")) {
		MOD_LOGD("SatelliteChanged");
		satellite_callback(param, user_data);
	}
	else if (!g_strcmp0(sig, "PositionChanged")) {
		MOD_LOGD("PositionChanged");
		position_callback(param, user_data);
	}
	else if (!g_strcmp0(sig, "StatusChanged")) {
		MOD_LOGD("StatusChanged");
		status_callback(param, user_data);
	}
	else {
		MOD_LOGD("Invaild signal[%s]", sig);
	}
}

static int start(gpointer handle, LocModStatusCB status_cb, LocModPositionExtCB pos_cb, LocModSatelliteCB sat_cb, gpointer userdata)
{
	MOD_LOGD("start");
	GpsManagerData *gps_manager = (GpsManagerData *) handle;
	g_return_val_if_fail(gps_manager, LOCATION_ERROR_NOT_AVAILABLE);
	g_return_val_if_fail(status_cb, LOCATION_ERROR_NOT_AVAILABLE);
	g_return_val_if_fail(pos_cb, LOCATION_ERROR_NOT_AVAILABLE);

	gps_manager->status_cb = status_cb;
	gps_manager->pos_cb = pos_cb;
	gps_manager->sat_cb = sat_cb;
	gps_manager->userdata = userdata;

	int ret = LBS_CLIENT_ERROR_NONE;
	ret = lbs_client_create (GPS_MANAGER_SERVICE_NAME, GPS_MANAGER_SERVICE_PATH, &(gps_manager->lbs_client));
	if (ret != LBS_CLIENT_ERROR_NONE || !gps_manager->lbs_client) {
		MOD_LOGE("Fail to create lbs_client_h. Error[%d]", ret);
		return LOCATION_ERROR_NOT_AVAILABLE;
	}
	MOD_LOGD("gps-manger(%x)", gps_manager);
	MOD_LOGD("pos_cb (%x), user_data(%x)", gps_manager->pos_cb, gps_manager->userdata);

	ret = lbs_client_start (gps_manager->lbs_client, LBS_CLIENT_LOCATION_CB | LBS_CLIENT_LOCATION_STATUS_CB | LBS_CLIENT_SATELLITE_CB, on_signal_callback, gps_manager);
	if (ret != LBS_CLIENT_ERROR_NONE) {
		MOD_LOGE("Fail to start lbs_client_h. Error[%d]", ret);
		return LOCATION_ERROR_NOT_AVAILABLE;
	}

	return LOCATION_ERROR_NONE;
}

static int stop(gpointer handle)
{
	MOD_LOGD("stop");
	GpsManagerData *gps_manager = (GpsManagerData *) handle;
	g_return_val_if_fail(gps_manager, LOCATION_ERROR_NOT_AVAILABLE);
	g_return_val_if_fail(gps_manager->lbs_client, LOCATION_ERROR_NOT_AVAILABLE);
	g_return_val_if_fail(gps_manager->status_cb, LOCATION_ERROR_NOT_AVAILABLE);

	int ret = LBS_CLIENT_ERROR_NONE;

	ret = lbs_client_stop (gps_manager->lbs_client);
	if (ret != LBS_CLIENT_ERROR_NONE) {
		MOD_LOGE("Fail to stop. Error[%d]", ret);
		return LOCATION_ERROR_NOT_AVAILABLE;
	}

	lbs_client_destroy(gps_manager->lbs_client);
	gps_manager->lbs_client = NULL;

	if (gps_manager->status_cb) {
		gps_manager->status_cb (FALSE, LOCATION_STATUS_NO_FIX, gps_manager->userdata);
	}

	gps_manager->status_cb = NULL;
	gps_manager->pos_cb = NULL;
	gps_manager->sat_cb = NULL;

	return LOCATION_ERROR_NONE;
}

static int get_nmea(gpointer handle, gchar ** nmea_data)
{
	MOD_LOGD("get_nmea");
	GpsManagerData *gps_manager = (GpsManagerData *) handle;
	g_return_val_if_fail(gps_manager, LOCATION_ERROR_NOT_AVAILABLE);
	g_return_val_if_fail(nmea_data, LOCATION_ERROR_PARAMETER);

	gboolean ret = FALSE;
	int timestamp = 0;
	ret = lbs_client_get_nmea(gps_manager->lbs_client, &timestamp, nmea_data);
	if (ret) {
		MOD_LOGE("\t Error getting nmea: %d", ret);
		return LOCATION_ERROR_NOT_AVAILABLE;
	}

	return LOCATION_ERROR_NONE;
}

static int get_last_position(gpointer handle, LocationPosition ** position, LocationVelocity ** velocity, LocationAccuracy ** accuracy)
{
	MOD_LOGD("get_last_position");
	GpsManagerData *gps_manager = (GpsManagerData *) handle;
	g_return_val_if_fail(gps_manager, LOCATION_ERROR_NOT_AVAILABLE);
	g_return_val_if_fail(position, LOCATION_ERROR_PARAMETER);
	g_return_val_if_fail(accuracy, LOCATION_ERROR_PARAMETER);

	int timestamp = 0;
	double longitude = 0.0, latitude = 0.0, altitude = 0.0;
	double speed = 0.0, direction = 0.0;
	double hor_accuracy = 0.0, ver_accuracy = 0.0;
	LocationStatus status = LOCATION_STATUS_NO_FIX;
	LocationAccuracyLevel level = LOCATION_ACCURACY_LEVEL_NONE;

	if (vconf_get_int(VCONFKEY_LOCATION_LAST_GPS_TIMESTAMP, &timestamp) ||
		vconf_get_dbl(VCONFKEY_LOCATION_LAST_GPS_LATITUDE, &latitude) ||
		vconf_get_dbl(VCONFKEY_LOCATION_LAST_GPS_LONGITUDE, &longitude) ||
		vconf_get_dbl(VCONFKEY_LOCATION_LAST_GPS_ALTITUDE, &altitude) ||
		vconf_get_dbl(VCONFKEY_LOCATION_LAST_GPS_SPEED, &speed) ||
		vconf_get_dbl(VCONFKEY_LOCATION_LAST_GPS_DIRECTION, &direction) ||
		vconf_get_dbl(VCONFKEY_LOCATION_LAST_GPS_HOR_ACCURACY, &hor_accuracy) ||
		vconf_get_dbl(VCONFKEY_LOCATION_LAST_GPS_VER_ACCURACY, &ver_accuracy)) {
		*position = NULL;
		*accuracy = NULL;
		return LOCATION_ERROR_NOT_AVAILABLE;
	}

	if (timestamp) {
		if (altitude) status = LOCATION_STATUS_3D_FIX;
		else status = LOCATION_STATUS_2D_FIX;
	}
	else {
		*position = NULL;
		*accuracy = NULL;
		return LOCATION_ERROR_NOT_AVAILABLE;
	}

	level = LOCATION_ACCURACY_LEVEL_DETAILED;
	*position = location_position_new(timestamp, latitude, longitude, altitude, status);
	*velocity = location_velocity_new ((guint) timestamp, speed * MPS_TO_KMPH, direction, 0.0);
	*accuracy = location_accuracy_new(level, hor_accuracy, ver_accuracy);

	return LOCATION_ERROR_NONE;
}

static int set_devname(gpointer handle, const char *devname)
{
	GpsManagerData *gps_manager = (GpsManagerData *) handle;
	g_return_val_if_fail(gps_manager, LOCATION_ERROR_NOT_AVAILABLE);
	g_return_val_if_fail(gps_manager->devname, LOCATION_ERROR_NOT_AVAILABLE);
	g_return_val_if_fail(devname, LOCATION_ERROR_PARAMETER);
	MOD_LOGD("set_devname: %s --> %s", gps_manager->devname, devname);
	g_stpcpy(gps_manager->devname, devname);

	return LOCATION_ERROR_NONE;
}

static int get_devname(gpointer handle, char **devname)
{
	GpsManagerData *gps_manager = (GpsManagerData *) handle;
	g_return_val_if_fail(gps_manager, LOCATION_ERROR_NOT_AVAILABLE);
	g_return_val_if_fail(gps_manager->devname, LOCATION_ERROR_NOT_AVAILABLE);
	g_return_val_if_fail(devname, LOCATION_ERROR_PARAMETER);
	*devname = g_strdup(gps_manager->devname);
	MOD_LOGD("get_devname: %s", *devname);

	return LOCATION_ERROR_NONE;
}

LOCATION_MODULE_API gpointer init(LocModGpsOps * ops)
{
	MOD_LOGD("init");

	g_return_val_if_fail(ops, NULL);
	ops->start = start;
	ops->stop = stop;
	ops->get_last_position = get_last_position;

	Dl_info info;
	if (dladdr(&get_last_position, &info) == 0) {
		MOD_LOGD("Failed to get module name");
	} else if (g_strrstr(info.dli_fname, "gps")) {
		ops->get_nmea = get_nmea;
		ops->set_devname = set_devname;
		ops->get_devname = get_devname;
	}

	GpsManagerData *gps_manager = g_new0(GpsManagerData, 1);
	g_return_val_if_fail(gps_manager, NULL);

	g_stpcpy(gps_manager->devname, "/dev/ttySAC1");
	gps_manager->status_cb = NULL;
	gps_manager->pos_cb = NULL;
	gps_manager->userdata = NULL;
	gps_manager->is_started = FALSE;

	return (gpointer) gps_manager;
}

LOCATION_MODULE_API void shutdown(gpointer handle)
{
	MOD_LOGD("shutdown");
	g_return_if_fail(handle);
	GpsManagerData *gps_manager = (GpsManagerData *) handle;
	if (gps_manager->lbs_client) {
		lbs_client_destroy(gps_manager->lbs_client);
		gps_manager->lbs_client = NULL;
	}

	g_free(gps_manager);

}
